/* Copyright 2006 Jetro Lauha - http://iki.fi/jetro/
 *
 * $Id: GameState.cpp,v 1.3 2006-04-30 22:11:53 tonic Exp $
 * $Revision: 1.3 $
 */

#include <stdio.h>
#include <time.h>
#include <math.h>

#include "main.h"
#include "GameState.h"


#define HELP_TEXT_COUNT 19
static char *sHelpTexts[HELP_TEXT_COUNT] = {
    "Down here will be valuable tips, read these junior warriors!",
    "Click screen to set target for your ship. Target is shown as a green light.",
    "Inactive collector has a white light and active one has a red light.",
    "Go near a collector to grab or activate it.",
    "Pull a collector near a bunch of particle elements and activate it.",
    "Activated collector gathers particle elements automatically.",
    "Disruptor base has blue lights.",
    "Bring particle elements to disruptor base by dragging active collectors.",
    "Amount of power in collectors is limited.",
    "Collector with low power starts to rotate slowly clockwise.",
    "Bring collector with gathered elements to base before it runs out of power.",
    "Try to gather as many particle elements as you can.",
    "Disruptor base gets power only when having an active collector near it.",
    "Fly carefully when you are dragging a collector.",
    "Level ends when all collectors run out of power.",
    "You can press space to center view on ship, B to center view on base.",
    "And you can press C repeatedly to cycle view through collectors.",
    "That's about it, you are now graduated as one of the Lone Warriors!",
    "Accept extra points for graduation by pressing Q."
};


/*
// for some reason this thing was slower for me than normal 1/sqrt(..)

// Inverse square root
// http://www.math.purdue.edu/~clomont/Math/Papers/2003/InvSqrt.pdf
static __forceinline float sInvSqrt(float x)
{
    float xhalf = 0.5f * x;
    int i = *(int *)&x; // get bits for floating value
    i = 0x5f375a86 - (i >> 1); // gives initial guess y0
    x = *(float *)&i; // convert bits back to float
    x = x * (1.5f - xhalf * x * x); // Newton step, repeating increases accuracy
    return x;
}
*/


static FLOAT step(FLOAT value, FLOAT stepPosition)
{
    return value < stepPosition ? 0 : 1;
}

static FLOAT smoothStep(FLOAT value, FLOAT slopeStart, FLOAT slopeEnd)
{
    FLOAT diff;
    if (value < slopeStart)
        return 0;
    if (value >= slopeEnd)
        return 1;

    diff = slopeEnd - slopeStart;
    if (diff == 0)
        return step(value, slopeStart);

    value = (value - slopeStart) / diff;     // normalize to [0..1]
    return value * value * (3 - 2 * value);
}

static FLOAT smoothPulse(FLOAT value, FLOAT upSlopeStart, FLOAT upSlopeEnd,
                         FLOAT downSlopeStart, FLOAT downSlopeEnd)
{
    return smoothStep(value, upSlopeStart, upSlopeEnd) -
           smoothStep(value, downSlopeStart, downSlopeEnd);
}


GameState::GameState(UIGraphics *graphics) :
    mTime(0), mStartTime(0),
    mGraphics(graphics),
    mLevelLabel(NULL),
    mContinueButton(NULL), mGrabButton(NULL),
    mDetachButton(NULL), mActivateButton(NULL),
    mButtonUpImage(NULL), mButtonDownImage(NULL),
    mSmallButtonUpImage(NULL), mSmallButtonDownImage(NULL),
    mBackgroundImage(NULL),
    mCollectorImage(NULL), mFlareImage(NULL),
    mBaseImage(NULL), mFlameImage(NULL)
{
    INT a;
    for (a = 0; a < LEVEL_COUNT; ++a)
        mColorsImage[a] = NULL;
    for (a = 0; a < SHIP_IMAGE_COUNT; ++a)
        mShipImages[a] = NULL;
}


GameState::~GameState()
{
}


BOOL GameState::init()
{
    BOOL success = TRUE;

    GameApp *ga = GameApp::getInstance();
    INT scrW = ga->getScreenWidth(), scrH = ga->getScreenHeight();
    UIGraphics *graphics = ga->getGraphics();
    mUI.setGraphics(graphics);

    if (success)
    {
        mButtonUpImage = graphics->loadImage(DATA_PREFIX "button_up.png");
        mButtonDownImage = graphics->loadImage(DATA_PREFIX "button_down.png");
        mSmallButtonUpImage = graphics->loadImage(DATA_PREFIX "button_up_small.png", UIImage::FILTER_LINEAR);
        mSmallButtonDownImage = graphics->loadImage(DATA_PREFIX "button_down_small.png", UIImage::FILTER_LINEAR);
        if (!mButtonDownImage || !mButtonDownImage ||
            !mSmallButtonDownImage || !mSmallButtonDownImage)
            success = FALSE;
    }

    if (success)
    {
        mBackgroundImage = graphics->loadImage(DATA_PREFIX "bkg1.png", UIImage::FILTER_LINEAR);
        mCollectorImage = graphics->loadImage(DATA_PREFIX "collector.png", UIImage::FILTER_LINEAR);
        mFlareImage = graphics->loadImage(DATA_PREFIX "flare.png", UIImage::FILTER_LINEAR);
        mBaseImage = graphics->loadImage(DATA_PREFIX "base.png", UIImage::FILTER_LINEAR);
        mFlameImage = graphics->loadImage(DATA_PREFIX "flame.png", UIImage::FILTER_LINEAR);
        if (!mBackgroundImage || !mCollectorImage ||
            !mFlareImage || !mBaseImage || !mFlameImage)
            success = FALSE;
    }

    if (success)
    {
        INT a;
        for (a = 0; a < LEVEL_COUNT && success; ++a)
        {
            char buf[1024];
            sprintf(buf, DATA_PREFIX "bkg_colors%02d.png", a + 1);
            mColorsImage[a] = graphics->loadImage(buf, UIImage::FILTER_LINEAR);
            if (!mColorsImage[a])
                success = FALSE;
        }
    }

    if (success)
    {
        INT a;
        for (a = 0; a < SHIP_IMAGE_COUNT && success; ++a)
        {
            char buf[1024];
            sprintf(buf, DATA_PREFIX "ship%04d.png", a);
            mShipImages[a] = graphics->loadImage(buf, UIImage::FILTER_LINEAR);
            if (!mShipImages[a])
                success = FALSE;
        }
    }

    if (success)
    {
        UIFont *medium = ga->getMediumFont();
        UIFont *small = ga->getSmallFont();
        UIFont *big = ga->getBigFont();
        UIFont *block = ga->getBlockFont();
        const UINT buttonWidth = mButtonUpImage->getWidth();
        const UINT buttonHeight = mButtonUpImage->getHeight();

        mLevelLabel = new UILabel(WIDTH / 2, HEIGHT / 2, NULL, big, "");
        mUI.addUIComponent(mLevelLabel);

        mContinueButton = new UIButton(scrW / 2 - buttonWidth / 2, scrH * 3 / 4,
                                       mButtonUpImage, mButtonDownImage, NULL,
                                       medium, "Continue");
        mContinueButton->addEventListener(this);
        mUI.addUIComponent(mContinueButton);
        if (!mContinueButton)
            success = FALSE;

        mGrabButton = new UIButton(100, 100,    // x,y dynamically updated
            mSmallButtonUpImage, mSmallButtonDownImage, NULL,
            block, "Grab");
        mGrabButton->addEventListener(this);
        mUI.addUIComponent(mGrabButton);
        if (!mGrabButton)
            success = FALSE;

        mDetachButton = new UIButton(100, 100,    // x,y dynamically updated
            mSmallButtonUpImage, mSmallButtonDownImage, NULL,
            block, "Detach");
        mDetachButton->addEventListener(this);
        mUI.addUIComponent(mDetachButton);
        if (!mDetachButton)
            success = FALSE;

        mActivateButton = new UIButton(100, 100,    // x,y dynamically updated
            mSmallButtonUpImage, mSmallButtonDownImage, NULL,
            block, "Activate");
        mActivateButton->addEventListener(this);
        mUI.addUIComponent(mActivateButton);
        if (!mActivateButton)
            success = FALSE;

        //mUI.activateAllComponents();
    }

    return success;
}


void GameState::deinit()
{
    mUI.deleteAllUIComponents();

    delete mButtonUpImage;
    delete mButtonDownImage;
    delete mSmallButtonUpImage;
    delete mSmallButtonDownImage;
    delete mBackgroundImage;
    delete mCollectorImage;
    delete mFlareImage;
    delete mBaseImage;
    delete mFlameImage;
    INT a;
    for (a = 0; a < LEVEL_COUNT; ++a)
        delete mColorsImage[a];
    for (a = 0; a < SHIP_IMAGE_COUNT; ++a)
        delete mShipImages[a];
}


void GameState::onActivate(UINT32 /*time*/)
{
    // TODO/FIXME
    //setMusic(TRUE, MUSIC_BASE);

    resetGame();
}


void GameState::resetGame()
{
    mUsedGraduationExtraPoints = FALSE;
    mHelpTextStartTick = 0;
    mTime = 0;
    mStartTime = 0;
    mScore = 0;
    mLevel = 1;
    mRandom.init_genrand(time(NULL));

    startLevel(mLevel);
}


void GameState::startLevel(INT level)
{
    INT32 a;
    char buf[1024];

    mLastTargetSoundTick = 0;

    mBaseSoundVolume = 0;
    setSoundChannelVolume(BASE_SND_CHANNEL, mBaseSoundVolume);
    mBaseSoundFrequency = 20000;
    setSoundChannelFrequency(BASE_SND_CHANNEL, mBaseSoundFrequency);

    sprintf(buf, "Level %d", level);
    mLevelLabel->setText(buf);

    mUI.activateUIComponent(mLevelLabel);
    mUI.activateUIComponent(mContinueButton);
    mBackgroundAngle = 0;
    mState = STATE_BRIEFING;
    mCollectorEnergyLeft = TRUE;

    mNextCollectorIndex = 0;
    mNextViewCenterCollectorIndex = 0;
    /*
    centerMouse();
    */
    mViewCenterX = mViewCenterY = 0;
    mViewCenterTargetX = mViewCenterTargetY = 0;

    mCollectors.clear();
    mActiveCollectors.clear();
    mParticles.clear();

    mVisualScale = VISUAL_SCALE_BASE / 2.25f;
    mParticlesLeft = 0;


    assert(level > 0);

    switch (level)
    {
    case 1:
        {
            mShip.reset(0.1f, -0.3f);
            mBase.reset(-0.5f, -0.5f);
            mWorldRadius = 0.7f;
            mParticlesLeft = 500;

            Collector c(0.5f, -0.6f, 0, COLLECTOR_NORMAL_AGE * 3 / 4, mNextCollectorIndex++);
            mCollectors.push_back(c);
        }
        break;

    case 2:
        {
            mShip.reset(0.4f, -0.3f);
            mBase.reset(0.5f, 0.2f);
            mWorldRadius = 1;
            mParticlesLeft = 3000;

            Collector c1(-0.5f, -0.5f, 0, COLLECTOR_NORMAL_AGE, mNextCollectorIndex++);
            mCollectors.push_back(c1);
            Collector c2(0, 0.3f, 0, COLLECTOR_NORMAL_AGE, mNextCollectorIndex++);
            mCollectors.push_back(c2);
        }
        break;

    case 3:
        {
            mShip.reset(0.3f, 0);
            mBase.reset(0, 0);
            mWorldRadius = 1.3f;
            mParticlesLeft = 5000;

            Collector c1(-0.1f, -0.25f, mRandom.genrand_real1(), COLLECTOR_NORMAL_AGE, mNextCollectorIndex++);
            mCollectors.push_back(c1);
            Collector c2(0.1f, -0.2f, mRandom.genrand_real1(), COLLECTOR_NORMAL_AGE, mNextCollectorIndex++);
            mCollectors.push_back(c2);
            Collector c3(0.3f, 0.25f, mRandom.genrand_real1(), COLLECTOR_NORMAL_AGE, mNextCollectorIndex++);
            mCollectors.push_back(c3);
        }
        break;
    }

    assert(mParticlesLeft > 0);

    for (a = 0; a < mParticlesLeft; ++a)
    {
        FLOAT x, y;
        /*
        x = (mRandom.genrand_real2() - 0.5f);// * ratio;
        y = (mRandom.genrand_real2() - 0.5f);
        */

        switch (level)
        {
        case 1:
            {
                // uniform angle, bias for center (but no units in immediate center)
                FLOAT angle = mRandom.genrand_real2() * PI * 2;
                FLOAT dist = mRandom.genrand_real1() * 0.5f + 0.25f;
                dist *= dist;
                x = (FLOAT)cos(angle) * dist;
                y = (FLOAT)sin(angle) * dist;
            }
            break;

        case 2:
            {
                // spiral legs
                const INT sectors = 4;
                FLOAT angle = (mRandom.genrand_int31() % sectors + mRandom.genrand_real2() / 2) / sectors * PI * 2;
                //float angle = mRandom.genrand_real2() * PI * 2;
                FLOAT dist = mRandom.genrand_real1() + 0.32f;
                //angle += dist * 4;
                dist *= dist;
                x = (FLOAT)cos(angle) * dist;
                y = (FLOAT)sin(angle) * dist;
                x += (mRandom.genrand_real1() - 0.5f) * 0.1f;
                y += (mRandom.genrand_real1() - 0.5f) * 0.1f;
            }
            break;

        case 3:
            {
                // spiral legs
                const INT sectors = 4;
                FLOAT angle = (mRandom.genrand_int31() % sectors + mRandom.genrand_real2() / 2) / sectors * PI * 2;
                //float angle = mRandom.genrand_real2() * PI * 2;
                FLOAT dist = mRandom.genrand_real1() + 0.32f;
                angle += dist * 4;
                dist *= dist;
                x = (FLOAT)cos(angle) * dist;
                y = (FLOAT)sin(angle) * dist;
                x += (mRandom.genrand_real1() - 0.5f) * 0.1f;
                y += (mRandom.genrand_real1() - 0.5f) * 0.1f;
            }
            break;
        }

        float vx = (mRandom.genrand_real1() - 0.5f) * 0.00001f;
        float vy = (mRandom.genrand_real1() - 0.5f) * 0.00001f;
        Particle p(x, y, vx, vy);
        mParticles.push_back(p);
    }
}

BOOL GameState::update(UINT32 time)
{
    if (mStartTime == 0)
        mStartTime = time;
    mTime = time - mStartTime;

    updateView();

    updateCollectors();

    //INT a; for (a = 0; a < 5; ++a)
        simulateParticles();

    updateShip();

    if (mBase.active)
        mBase.update();

    if (mParticlesLeft <= 0 ||
        !mCollectorEnergyLeft)
    {
        if (mState != STATE_FINISHED)
            playSoundEffect(SOUND_LEVELDONE);
        mState = STATE_FINISHED;
    }

    mBaseSoundVolume -= 1;
    if (mBaseSoundVolume < 0)
        mBaseSoundVolume = 0;
    mBaseSoundFrequency *= 0.99f;
    if (mBaseSoundFrequency < 20000)
        mBaseSoundFrequency = 20000;
    setSoundChannelVolume(BASE_SND_CHANNEL, (INT32)mBaseSoundVolume);
    setSoundChannelFrequency(BASE_SND_CHANNEL, (INT32)mBaseSoundFrequency);

    return TRUE;
}


void GameState::updateCollectors()
{
    INT a;
    mCollectorEnergyLeft = FALSE;
    mBase.active = FALSE;
    for (a = 0; a < mCollectors.size(); ++a)
    {
        mCollectors[a].update();

        if (mCollectors[a].runTimeLeft > 0)
            mCollectorEnergyLeft = TRUE;

        if (mCollectors[a].runTimeLeft <= 0 && mCollectors[a].active)
        {
            mCollectors[a].active = FALSE;
            CollectorPointerArray::iterator it, itEnd;
            itEnd = mActiveCollectors.end();
            for (it = mActiveCollectors.begin(); it != itEnd; ++it)
            {
                if (*it == &mCollectors[a])
                {
                    mActiveCollectors.erase(it);
                    break;
                }
            }
        }

        if (mCollectors[a].active)
        {
            FLOAT dx = mCollectors[a].x - mBase.x;
            FLOAT dy = mCollectors[a].y - mBase.y;
            FLOAT dsq = dx * dx + dy * dy;
            if (dsq < BASE_ACTIVE_RADIUS * BASE_ACTIVE_RADIUS)
                mBase.active = TRUE;
        }
    }
}


void GameState::updateView()
{
    if (mState != STATE_BRIEFING && mVisualScale < VISUAL_SCALE_BASE)
        mVisualScale = mVisualScale * 0.99 + VISUAL_SCALE_BASE * 0.01;

    INT32 mx = mMouseX - WIDTH / 2;
    INT32 my = mMouseY - HEIGHT / 2;
    if (mState == STATE_RUNNING &&
        mx * mx + my * my > (HEIGHT / 6) * (HEIGHT / 6))
    {
        FLOAT dirX = mx * 0.00003f;
        FLOAT dirY = my * 0.00003f;
        //mViewCenterTargetX = mViewCenterX + dirX;
        //mViewCenterTargetY = mViewCenterY + dirY;
        mViewCenterTargetX += dirX;
        mViewCenterTargetY += dirY;
        FLOAT viewCenterDistSq = mViewCenterTargetX * mViewCenterTargetX + mViewCenterTargetY * mViewCenterTargetY;
        if (viewCenterDistSq > mWorldRadius * mWorldRadius)
        {
            FLOAT viewCenterDistLength = (FLOAT)sqrt(viewCenterDistSq);
            mViewCenterTargetX = mWorldRadius * (mViewCenterTargetX / viewCenterDistLength);
            mViewCenterTargetY = mWorldRadius * (mViewCenterTargetY / viewCenterDistLength);
        }
    }

    if (mState == STATE_BRIEFING)
    {
        mViewCenterTargetX = mViewCenterTargetY = 0;
        mBackgroundAngle = 0;
    }
    else if (mState == STATE_RUNNING)
    {
        mBackgroundAngle += 0.00005f;
        mUI.deactivateUIComponent(mContinueButton);
        mUI.deactivateUIComponent(mLevelLabel);
    }
    else if (mState == STATE_FINISHED)
    {
        mBackgroundAngle += 0.00005f;
        mUI.activateUIComponent(mContinueButton);
    }

    mViewCenterX = mViewCenterX * 0.97f + mViewCenterTargetX * 0.03f;
    mViewCenterY = mViewCenterY * 0.97f + mViewCenterTargetY * 0.03f;
}


void GameState::simulateParticles()
{
    ParticleArray::iterator it = mParticles.begin(), itEnd = mParticles.end();
    const INT collectorCount = mActiveCollectors.size();
    Particle *p;
    const FLOAT damping = DAMPING;
    for (; it != itEnd; /*p->vx *= damping, p->vy *= damping,*/ ++it)
    {
        //damping = DAMPING;
        p = &*it;
        if (!p->active)
            continue;
        p->x += p->vx;
        p->y += p->vy;

        p->vx *= damping;
        p->vy *= damping;

        INT a;
        for (a = 0; a < collectorCount; ++a)
        {
            Collector *c = mActiveCollectors[a];
            const FLOAT targetX = c->x;
            const FLOAT targetY = c->y;
            const FLOAT boost = 1.01f;
            //const FLOAT boost = (a == mShip.attachedCollector) ? 1.1f : 1.01f;
            FLOAT tx = targetX - p->x;
            FLOAT ty = targetY - p->y;
            FLOAT dsq = tx * tx + ty * ty;
            if (dsq > COLLECTOR_GREATER_RADIUS * COLLECTOR_GREATER_RADIUS)
                continue;

            if (dsq > COLLECTOR_RADIUS * COLLECTOR_RADIUS)
            {
                FLOAT dsqInvSqrt = (FLOAT)(1 / sqrt(dsq));
                //FLOAT dsqInvSqrt = sInvSqrt(dsq);
                FLOAT vtx = tx * dsqInvSqrt;
                FLOAT vty = ty * dsqInvSqrt;
                p->vx += vtx * 0.0000001f;
                p->vy += vty * 0.0000001f;
            }
            else
            {
                if (/*mShip.goingToTarget &&*/
                    dsq < COLLECTOR_TOW_RADIUS * COLLECTOR_TOW_RADIUS &&
                    c->index == mShip.attachedCollector)
                {
                    //damping = DAMPING_ATTACHED;
                    //p->x += c->x - c->px;
                    //p->y += c->y - c->py;
                    p->x += c->dx;
                    p->y += c->dy;
                }

                FLOAT dsqInvSqrt = (FLOAT)(1 / sqrt(dsq));
                //FLOAT dsqInvSqrt = sInvSqrt(dsq);
                FLOAT vtx = tx * dsqInvSqrt;
                FLOAT vty = ty * dsqInvSqrt;
                FLOAT vtxn = -vty * 0.02f;
                FLOAT vtyn = vtx * 0.02f;
                if (dsq < (COLLECTOR_RADIUS / 4) * (COLLECTOR_RADIUS / 4)) {
                    // push particles away from direct center
                    vtx = -vtx;
                    vty = -vty;
                } else {
                    vtx *= boost;
                    vty *= boost;
                }
                //p->vx = (p->vx + vtx) * 0.005f;
                //p->vy = (p->vy + vty) * 0.005f;
                p->vx += (vtx + vtxn) * 0.000001f;
                p->vy += (vty + vtyn) * 0.000001f;
            }
        }

        // ship
        {
            FLOAT tx = p->x - mShip.x;
            FLOAT ty = p->y - mShip.y;
            FLOAT dsq = tx * tx + ty * ty;
            if (dsq < SHIP_RADIUS * SHIP_RADIUS)
            {
                FLOAT dsqInvSqrt = (FLOAT)(1 / sqrt(dsq));
                FLOAT vtx = tx * dsqInvSqrt;
                FLOAT vty = ty * dsqInvSqrt;
                FLOAT m = mShip.braking > 0 ? 0.00007f : 0.00003f;
                p->vx += vtx * m;
                p->vy += vty * m;
            }
        }

        // base
        if (mBase.active)
        {
            FLOAT tx = p->x - mBase.x;
            FLOAT ty = p->y - mBase.y;
            FLOAT dsq = tx * tx + ty * ty;
            if (dsq < BASE_RADIUS * BASE_RADIUS)
            {
                if (dsq < BASE_EAT_RADIUS * BASE_EAT_RADIUS)
                {
                    /* // uh, what was wrong with this? no time to look into it, making a workaround:
                    it = mParticles.erase(it); // returns next one
                    --it;   // for end statement does ++it which restores this to what it just was
                    */
                    p->active = FALSE;
                    mScore += 100;
                    --mParticlesLeft;

                    mBaseSoundVolume += 6;
                    if (mBaseSoundVolume > 255)
                        mBaseSoundVolume = 255;
                    mBaseSoundFrequency *= 1.07f;
                    if (mBaseSoundFrequency > 300000)
                        mBaseSoundFrequency = 300000;
                    setSoundChannelVolume(BASE_SND_CHANNEL, mBaseSoundVolume);
                    setSoundChannelFrequency(BASE_SND_CHANNEL, mBaseSoundFrequency);
                }
                else
                {
                    FLOAT dsqInvSqrt = (FLOAT)(1 / sqrt(dsq));
                    FLOAT vtx = tx * dsqInvSqrt;
                    FLOAT vty = ty * dsqInvSqrt;
                    const FLOAT m = 0.00005f;
                    p->vx -= vtx * m;
                    p->vy -= vty * m;
                }
            }
        }
    }
}


void GameState::updateShip()
{
    mShip.update(mTime);

    if (mShip.attachedCollector >= 0)
    {
        Collector &c = mCollectors[mShip.attachedCollector];
        FLOAT dx = c.x - mShip.x;
        FLOAT dy = c.y - mShip.y;
        FLOAT dsq = dx * dx + dy * dy;
        if (dsq > COLLECTOR_TOW_RADIUS * COLLECTOR_TOW_RADIUS)
        {
            FLOAT dsqInvSqrt = (FLOAT)(1 / sqrt(dsq));
            dx *= dsqInvSqrt;
            dy *= dsqInvSqrt;
            dx *= COLLECTOR_TOW_RADIUS;
            dy *= COLLECTOR_TOW_RADIUS;
            FLOAT ox = c.x, oy = c.y;
            //c.px = c.x;
            //c.py = c.y;
            c.x = c.x * 0.99f + (mShip.x + dx) * 0.01f;
            c.y = c.y * 0.99f + (mShip.y + dy) * 0.01f;
            c.dx = (c.x - ox) * PARTICLE_TOW_DISPLACE_RATIO;
            c.dy = (c.y - oy) * PARTICLE_TOW_DISPLACE_RATIO;
        }
    }
}


void GameState::updateContextButtons()
{
    INT a;
    FLOAT closestDistSq = 1e10f;
    INT closestCollector = -1;
    for (a = 0; a < mCollectors.size(); ++a)
    {
        FLOAT dx = mShip.x - mCollectors[a].x;
        FLOAT dy = mShip.y - mCollectors[a].y;
        FLOAT dsq = dx * dx + dy * dy;
        if (dsq < closestDistSq)
        {
            closestDistSq = dsq;
            closestCollector = a;
        }
    }

    mClosestCollector = closestCollector;

    if (closestCollector >= 0 &&
        closestDistSq < COLLECTOR_RADIUS * COLLECTOR_RADIUS)
    {
        if (mShip.attachedCollector >= 0)
            closestCollector = mShip.attachedCollector;

        FLOAT x = mCollectors[closestCollector].x * mVisualScale + WIDTH / 2 - mVisualScale * mViewCenterX;
        FLOAT y = mCollectors[closestCollector].y * mVisualScale + HEIGHT / 2 - mVisualScale * mViewCenterY;
        if (mShip.attachedCollector == closestCollector)
        {
            mDetachButton->setX(x - mSmallButtonUpImage->getWidth() / 2);
            mDetachButton->setY(y - mSmallButtonUpImage->getHeight() / 2 - 40);
            mUI.activateUIComponent(mDetachButton);
            mUI.deactivateUIComponent(mGrabButton);
        }
        else
        {
            mGrabButton->setX(x - mSmallButtonUpImage->getWidth() / 2);
            mGrabButton->setY(y - mSmallButtonUpImage->getHeight() / 2 + 40);
            mUI.activateUIComponent(mGrabButton);
            mUI.deactivateUIComponent(mDetachButton);
        }
        if (!mCollectors[closestCollector].active &&
            mCollectors[closestCollector].runTimeLeft > 0)
        {
            mActivateButton->setX(x - mSmallButtonUpImage->getWidth() / 2);
            mActivateButton->setY(y - mSmallButtonUpImage->getHeight() / 2);
            mUI.activateUIComponent(mActivateButton);
        }
        else
            mUI.deactivateUIComponent(mActivateButton);
    }
    else
    {
        mUI.deactivateUIComponent(mGrabButton);
        mUI.deactivateUIComponent(mDetachButton);
        mUI.deactivateUIComponent(mActivateButton);
    }
}


void GameState::render(SDL_Surface *screen)
{
    // clear with ambient color
    //glClearColor(0.15f, 0.15f, 0.15f, 1);
    //glClear(GL_COLOR_BUFFER_BIT);

    updateContextButtons();

    mGraphics->beginRender();

    drawBackground();

    drawCollectors();

    drawParticles();

    drawBase();

    drawCollectorLights();
    drawBaseLights();

    drawShip();

    drawTexts();

    // Testing..
    //GameApp::getInstance()->getBlockFont()->drawString(100, 100, 0, "Converter Base");

    mGraphics->endRender();

    mUI.render();
}


void GameState::drawBackground()
{
    const FLOAT backgroundAngle = mBackgroundAngle;
    const FLOAT backgroundScale = 3 * 1.17f * mVisualScale / VISUAL_SCALE_BASE;
    mBackgroundImage->setAngle(backgroundAngle);
    mBackgroundImage->setScale(backgroundScale);
    mBackgroundImage->draw(UIImage::NORMAL,
        UIImage::TRANSFORM | UIImage::ROTATE | UIImage::CENTER | UIImage::SCALE,
        WIDTH / 2 - mVisualScale * mViewCenterX / 2,
        HEIGHT / 2 - mVisualScale * mViewCenterY / 2,
        0.5f, 0.5f, 0.7f, 1);
    UIImage *i = mColorsImage[mLevel - 1];
    FLOAT backgroundColorScale = mBackgroundImage->getWidth() / i->getWidth();
    i->setAngle(backgroundAngle);
    i->setScale(backgroundScale * backgroundColorScale);
    i->draw(UIImage::MULTIPLY,
        UIImage::TRANSFORM | UIImage::ROTATE | UIImage::CENTER | UIImage::SCALE,
        WIDTH / 2 - mVisualScale * mViewCenterX / 2,
        HEIGHT / 2 - mVisualScale * mViewCenterY / 2);
}


void GameState::drawCollectors()
{
    INT a, collectorCount = mCollectors.size();
    FLOAT collectorImageScale = mVisualScale / VISUAL_SCALE_BASE;
    mCollectorImage->setScale(collectorImageScale);
    for (a = 0; a < collectorCount; ++a)
    {
        FLOAT x = mCollectors[a].x * mVisualScale + WIDTH / 2 - mVisualScale * mViewCenterX;
        FLOAT y = mCollectors[a].y * mVisualScale + HEIGHT / 2 - mVisualScale * mViewCenterY;
        mCollectorImage->setAngle(mCollectors[a].angle);
        mCollectorImage->draw(UIImage::NORMAL,
            UIImage::TRANSFORM | UIImage::ROTATE | UIImage::CENTER | UIImage::SCALE,
            x, y);
    }
}


void GameState::drawCollectorLights()
{
    INT a, collectorCount = mCollectors.size();
    FLOAT collectorImageScale = mVisualScale / VISUAL_SCALE_BASE;
    for (a = 0; a < collectorCount; ++a)
    {
        FLOAT x = mCollectors[a].x * mVisualScale + WIDTH / 2 - mVisualScale * mViewCenterX;
        FLOAT y = mCollectors[a].y * mVisualScale + HEIGHT / 2 - mVisualScale * mViewCenterY;
        FLOAT size = (1 - (FLOAT)sin(fmod(mTime * 0.001f + a, PI)));
        if (mCollectors[a].active)
        {
            size *= size * size * size * size;
            size = size * 2 + 0.1f;
            mFlareImage->setScale(collectorImageScale * size);
            mFlareImage->draw(UIImage::ADD,
                UIImage::TRANSFORM | UIImage::CENTER | UIImage::SCALE,
                x, y, 1, 0, 0, 1);
        }
        else
        {
            size *= size;
            size = size / 2 + 0.1f;
            mFlareImage->setScale(collectorImageScale * size);
            mFlareImage->draw(UIImage::ADD,
                UIImage::TRANSFORM | UIImage::CENTER | UIImage::SCALE,
                x, y, 1, 1, 1, 1);
        }
    }
}


void GameState::drawBase()
{
    FLOAT imageScale = mVisualScale / VISUAL_SCALE_BASE;
    mBaseImage->setScale(imageScale);
    FLOAT x = mBase.x * mVisualScale + WIDTH / 2 - mVisualScale * mViewCenterX;
    FLOAT y = mBase.y * mVisualScale + HEIGHT / 2 - mVisualScale * mViewCenterY;
    mBaseImage->setAngle(mBase.angle);
    mBaseImage->draw(UIImage::NORMAL,
        UIImage::TRANSFORM | UIImage::ROTATE | UIImage::CENTER | UIImage::SCALE,
        x, y);
}


void GameState::drawBaseLights()
{
    FLOAT imageScale = mVisualScale / VISUAL_SCALE_BASE;
    mBaseImage->setScale(imageScale);
    FLOAT cx = mBase.x * mVisualScale + WIDTH / 2 - mVisualScale * mViewCenterX;
    FLOAT cy = mBase.y * mVisualScale + HEIGHT / 2 - mVisualScale * mViewCenterY;
    INT a;
    for (a = 0; a < 8; ++a)
    {
        FLOAT x = cx + (FLOAT)cos(a * 2 * PI / 8 + mBase.angle) * (BASE_RADIUS + 0.01) * mVisualScale;
        FLOAT y = cy + (FLOAT)sin(a * 2 * PI / 8 + mBase.angle) * (BASE_RADIUS + 0.01) * mVisualScale;
        FLOAT size = (1 - (FLOAT)sin(fmod(mTime * 0.001f + a * PI / 8, PI)));
        size *= size * size * size * size;
        size = size * 2 + 0.2f;
        mFlareImage->setScale(imageScale * size);
        mFlareImage->draw(UIImage::ADD,
            UIImage::TRANSFORM | UIImage::CENTER | UIImage::SCALE,
            x, y, 0, 0, 1, 1);
        size = size / 8 + 0.2f;
        mFlareImage->setScale(imageScale * size);
        mFlareImage->draw(UIImage::ADD,
            UIImage::TRANSFORM | UIImage::CENTER | UIImage::SCALE,
            x, y, 1, 1, 1, 1);
    }
}


void GameState::drawParticles()
{
    glDisable(GL_TEXTURE_2D);
    glEnable(GL_POINT_SMOOTH);
    glHint(GL_POINT_SMOOTH_HINT, GL_FASTEST);

    glPointSize(6.0f);
    glColor4f(0, 0, 0, 0.5f);
    glPushMatrix();
    glTranslatef(WIDTH / 2, HEIGHT / 2, 0);
    glScalef(mVisualScale, mVisualScale, 1);
    glTranslatef(-mViewCenterX, -mViewCenterY, 0);
    glBegin(GL_POINTS);
    {
        ParticleArray::iterator it = mParticles.begin(), itEnd = mParticles.end();
        for (; it != itEnd; ++it)
        {
            Particle &p = *it;
            if (!p.active)
                continue;
            float x = p.x, y = p.y;
            glVertex2f(x, y);
        }
    }
    glEnd();

    glPointSize(2.0f);
    glColor4f(0.75f, 0.75f, 1, 1);
    glBegin(GL_POINTS);
    {
        ParticleArray::iterator it = mParticles.begin(), itEnd = mParticles.end();
        for (; it != itEnd; ++it)
        {
            Particle &p = *it;
            if (!p.active)
                continue;
            float x = p.x, y = p.y;
            glVertex2f(x, y);
        }
    }
    glEnd();

    glEnable(GL_TEXTURE_2D);

    glPopMatrix();
    glColor4f(1, 1, 1, 1);
}


void GameState::drawShip()
{
    if (mShip.attachedCollector >= 0)
    {
        // TODO: ATTACH LINE
        /*
        glDisable(GL_TEXTURE_2D);
        glEnable(GL_LINE_SMOOTH);
        glHint(GL_LINE_SMOOTH_HINT, GL_FASTEST);
        glLineWidth(3);
        glColor4f(1, 0.25f, 0.6f, 0.5f);

        glEnable(GL_TEXTURE_2D);
        */
    }

    FLOAT imageScale = mVisualScale / VISUAL_SCALE_BASE;
    //UIImage *image = mShipImages[(mTime / 100) % SHIP_IMAGE_COUNT];
    UIImage *image = mShipImages[(((INT32)(SHIP_IMAGE_COUNT * (-mShip.angle - PI / 2) / (2 * PI)) % SHIP_IMAGE_COUNT) + SHIP_IMAGE_COUNT) % SHIP_IMAGE_COUNT];
    image->setScale(imageScale * 1.2f);
    image->draw(UIImage::NORMAL,
        UIImage::TRANSFORM | UIImage::CENTER | UIImage::SCALE,
        (mShip.x * 0.999f) * mVisualScale + WIDTH / 2 - mVisualScale * mViewCenterX,
        (mShip.y * 0.999f) * mVisualScale + HEIGHT / 2 - mVisualScale * mViewCenterY, 0, 0, 0, 0.2f);
    FLOAT x = mShip.x * mVisualScale + WIDTH / 2 - mVisualScale * mViewCenterX;
    FLOAT y = mShip.y * mVisualScale + HEIGHT / 2 - mVisualScale * mViewCenterY;
    image->setScale(imageScale);
    if (mShip.braking > 0)
    {
        image->draw(UIImage::NORMAL,
            UIImage::TRANSFORM | UIImage::CENTER | UIImage::SCALE,
            x + (mRandom.genrand_real1() - 0.5f) * 4,
            y + (mRandom.genrand_real1() - 0.5f) * 4,
            1, 1, 1, 0.6f);
        image->draw(UIImage::NORMAL,
            UIImage::TRANSFORM | UIImage::CENTER | UIImage::SCALE,
            x + (mRandom.genrand_real1() - 0.5f) * 4,
            y + (mRandom.genrand_real1() - 0.5f) * 4,
            1, 1, 1, 0.6f);

        FLOAT alpha = (FLOAT)sin(mTime * 0.06) * 0.5f + 0.5f;
        mFlameImage->setScale(imageScale / 2);
        mFlameImage->setAngle(mShip.angle - PI / 2 + PI / 4);
        mFlameImage->draw(UIImage::ADD,
            UIImage::TRANSFORM | UIImage::CENTER | UIImage::ROTATE | UIImage::SCALE,
            x - (FLOAT)cos(mShip.angle - PI + PI / 4) * 0.025f * mVisualScale,
            y - (FLOAT)sin(mShip.angle - PI + PI / 4) * 0.025f * mVisualScale,
            0, 0.5f, 1, alpha);
        mFlameImage->setAngle(mShip.angle - PI / 2 - PI / 4);
        mFlameImage->draw(UIImage::ADD,
            UIImage::TRANSFORM | UIImage::CENTER | UIImage::ROTATE | UIImage::SCALE,
            x - (FLOAT)cos(mShip.angle - PI - PI / 4) * 0.025f * mVisualScale,
            y - (FLOAT)sin(mShip.angle - PI - PI / 4) * 0.025f * mVisualScale,
            0, 0.5f, 1, alpha);
    }
    else
    {
        image->draw(UIImage::NORMAL,
            UIImage::TRANSFORM | UIImage::CENTER | UIImage::SCALE,
            mShip.x * mVisualScale + WIDTH / 2 - mVisualScale * mViewCenterX,
            mShip.y * mVisualScale + HEIGHT / 2 - mVisualScale * mViewCenterY);
    }

    if (mShip.goingToTarget)
    {
        FLOAT x = mShip.targetX * mVisualScale + WIDTH / 2 - mVisualScale * mViewCenterX;
        FLOAT y = mShip.targetY * mVisualScale + HEIGHT / 2 - mVisualScale * mViewCenterY;
        FLOAT size = (FLOAT)sin(fmod(mTime * 0.0035f, PI)) * 0.5f + 1;
        mFlareImage->setScale(imageScale * size / 2);
        mFlareImage->draw(UIImage::ADD,
            UIImage::TRANSFORM | UIImage::CENTER | UIImage::SCALE,
            x, y, 0, 1, 0, 1);

        FLOAT alpha = (FLOAT)sin(mTime * 0.03) * 0.5f + 0.5f;
        FLOAT scale = (FLOAT)sin(mTime * 0.1) * 0.2f + 1;
        x = mShip.x * mVisualScale + WIDTH / 2 - mVisualScale * mViewCenterX;
        y = mShip.y * mVisualScale + HEIGHT / 2 - mVisualScale * mViewCenterY;
        mFlameImage->setAngle(mShip.angle + PI / 2);
        mFlameImage->setScale(imageScale * scale);
        mFlameImage->draw(UIImage::ADD,
            UIImage::TRANSFORM | UIImage::CENTER | UIImage::ROTATE | UIImage::SCALE,
            x - (FLOAT)cos(mShip.angle) * 0.035f * mVisualScale * scale,
            y - (FLOAT)sin(mShip.angle) * 0.035f * mVisualScale * scale,
            1, 1, 1, alpha);
    }
}


void GameState::drawTexts()
{
    INT a;
    FLOAT x, y;
    char buf[100];
    UIFont *font = GameApp::getInstance()->getBlockFont();
    UIFont *sf = GameApp::getInstance()->getSmallFont();
    UIFont *bf = GameApp::getInstance()->getBigFont();

    sprintf(buf, "%d", mScore);
    font->drawString(WIDTH - 4, 4, UIFont::ALIGN_RIGHT, buf);

    //GameApp::getInstance()->getBlockFont()->drawString(5, 100, 0, "Click screen to set target for your ship. Target is shown as a green light.");
    const INT32 helpTextFade = 2000;
    const INT32 helpTextPeriod = 12000;
    const INT32 currentHelpIndex = (mTime - mHelpTextStartTick) / helpTextPeriod;
    if (mHelpTextStartTick != 0 && currentHelpIndex < HELP_TEXT_COUNT)
    {
        FLOAT alpha = smoothPulse(mTime - mHelpTextStartTick,
            currentHelpIndex * helpTextPeriod,
            currentHelpIndex * helpTextPeriod + helpTextFade,
            (currentHelpIndex + 1) * helpTextPeriod - helpTextFade,
            (currentHelpIndex + 1) * helpTextPeriod);
        font->drawString(WIDTH / 2, HEIGHT - 8,
            UIFont::ALIGN_BOTTOM | UIFont::ALIGN_HORIZ_CENTER,
            sHelpTexts[currentHelpIndex], 0.75f, 0.8f, 0.85f, alpha);
    }

    if (mState == STATE_BRIEFING)
    {
        x = mShip.x * mVisualScale + WIDTH / 2 - mVisualScale * mViewCenterX;
        y = mShip.y * mVisualScale + HEIGHT / 2 - mVisualScale * mViewCenterY;
        sf->drawString(x + 10, y, UIFont::ALIGN_VERT_CENTER, "SHIP", 0.75f, 1, 0.75f, 0.75f);

        for (a = 0; a < mCollectors.size(); ++a)
        {
            Collector &c = mCollectors[a];
            x = c.x * mVisualScale + WIDTH / 2 - mVisualScale * mViewCenterX;
            y = c.y * mVisualScale + HEIGHT / 2 - mVisualScale * mViewCenterY;
            sf->drawString(x, y, UIFont::ALIGN_HORIZ_CENTER | UIFont::ALIGN_VERT_CENTER, "COLLECTOR", 0.75f, 1, 0.75f, 0.75f);
        }

        x = mBase.x * mVisualScale + WIDTH / 2 - mVisualScale * mViewCenterX;
        y = mBase.y * mVisualScale + HEIGHT / 2 - mVisualScale * mViewCenterY;
        sf->drawString(x, y, UIFont::ALIGN_HORIZ_CENTER | UIFont::ALIGN_VERT_CENTER, "BASE", 0.75f, 1, 0.75f, 0.75f);
    }

    if (mState == STATE_FINISHED)
    {
        if (mLevel < LEVEL_COUNT)
        {
            bf->drawString(WIDTH / 2, HEIGHT / 2,
                UIFont::ALIGN_HORIZ_CENTER | UIFont::ALIGN_VERT_CENTER,
                "Level Done!");
        }
        else
        {
            bf->drawString(WIDTH / 2, HEIGHT / 3,
                UIFont::ALIGN_HORIZ_CENTER | UIFont::ALIGN_VERT_CENTER,
                "Game Finished!");
            font->drawString(WIDTH / 2, HEIGHT / 3 + bf->getFontHeight() + 20,
                UIFont::ALIGN_HORIZ_CENTER | UIFont::ALIGN_VERT_CENTER,
                "You're now a legendary Lone Warrior!");
            if (GameApp::getInstance()->getBestScore() < mScore)
            {
                char buf[1024];
                sprintf(buf, "NEW HIGHSCORE: %d", mScore);
                font->drawString(WIDTH / 2, HEIGHT / 3 + bf->getFontHeight() * 2 + 50,
                    UIFont::ALIGN_HORIZ_CENTER | UIFont::ALIGN_VERT_CENTER, buf);
            }
            else
            {
                char buf[1024];
                sprintf(buf, "Your score: %d", mScore);
                font->drawString(WIDTH / 2, HEIGHT / 3 + bf->getFontHeight() * 2 + 50,
                    UIFont::ALIGN_HORIZ_CENTER | UIFont::ALIGN_VERT_CENTER, buf);
            }
        }
    }
}


void GameState::centerMouse()
{
    mMouseX = WIDTH / 2;
    mMouseY = HEIGHT / 2;
    SDL_WarpMouse(mMouseX, mMouseY);
}


void GameState::onKeyDown(const SDLKey &key)
{
    if (key == SDLK_SPACE)
    {
        mViewCenterTargetX = mShip.x;
        mViewCenterTargetY = mShip.y;
        centerMouse();
    }
    else if (key == SDLK_b)
    {
        mViewCenterTargetX = mBase.x;
        mViewCenterTargetY = mBase.y;
        centerMouse();
    }
    else if (key == SDLK_c)
    {
        mViewCenterTargetX = mCollectors[mNextViewCenterCollectorIndex].x;
        mViewCenterTargetY = mCollectors[mNextViewCenterCollectorIndex].y;
        centerMouse();
        mNextViewCenterCollectorIndex = (mNextViewCenterCollectorIndex + 1) % mCollectors.size();
    }
    else if (key == SDLK_q && !mUsedGraduationExtraPoints)
    {
        mScore += 10000;
        mUsedGraduationExtraPoints = TRUE;
    }
    else if (key == SDLK_ESCAPE)
    {
        returnToTitleMenu();
    }
}


void GameState::onMouseDown(const SDL_MouseButtonEvent &event)
{
    if (mUI.handleUIPointerDown(event.x, event.y))
        return;

    if (mState == STATE_FINISHED || mState == STATE_BRIEFING)
        return;

    mMouseDown = TRUE;
    mMouseDownX = event.x;
    mMouseDownY = event.y;
}


void GameState::onMouseUp(const SDL_MouseButtonEvent &event)
{
    if (mUI.handleUIPointerUp(event.x, event.y))
        return;

    if (mState == STATE_FINISHED || mState == STATE_BRIEFING)
        return;

    mMouseDown = FALSE;

    INT x = event.x, y = event.y;
    if (abs(mMouseDownX - x) < 30 && abs(mMouseDownY - y) < 30)
    {
        if (mLastTargetSoundTick + 1000 < mTime)
        {
            playSoundEffect(SOUND_TARGET);
            mLastTargetSoundTick = mTime;
        }
        mShip.setTarget((FLOAT)(x - WIDTH / 2) / mVisualScale + mViewCenterX,
                        (FLOAT)(y - HEIGHT / 2) / mVisualScale + mViewCenterY);
    }
}


void GameState::onMouseMotion(const SDL_MouseMotionEvent &event)
{
    if (mUI.handleUIPointerMove(event.x, event.y))
        return;

    mMouseX = event.x;
    mMouseY = event.y;
}


void GameState::returnToTitleMenu()
{
    GameApp *ga = GameApp::getInstance();
    if (ga->getBestScore() < mScore)
        ga->setBestScore(mScore);
    ga->setState(GameApp::STATE_TITLE);
}


void GameState::onAction(UIComponent *component)
{
    if (component == mContinueButton)
    {
        if (mState == STATE_BRIEFING)
        {
            mState = STATE_RUNNING;
            mViewCenterTargetX = mShip.x;
            mViewCenterTargetY = mShip.y;
            centerMouse();
            if (mLevel == 1)
                mHelpTextStartTick = mTime;
            playSoundEffect(SOUND_STARTLEVEL);
        }
        else if (mState == STATE_FINISHED)
        {
            playSoundEffect(SOUND_BUTTON);
            if (mLevel < LEVEL_COUNT)
                startLevel(++mLevel);
            else
            {
                returnToTitleMenu();
            }
        }
    }
    else if (component == mGrabButton)
    {
        assert(mClosestCollector >= 0);
        if (mClosestCollector < 0) return;
        mShip.attachCollector(mClosestCollector);
        playSoundEffect(SOUND_GRAB);
    }
    else if (component == mDetachButton)
    {
        assert(mClosestCollector >= 0);
        if (mClosestCollector < 0) return;
        mShip.detachCollector();
        playSoundEffect(SOUND_DETACH);
    }
    else if (component == mActivateButton)
    {
        assert(mClosestCollector >= 0);
        if (mClosestCollector < 0) return;
        if (!mCollectors[mClosestCollector].active &&
            mCollectors[mClosestCollector].runTimeLeft > 0)
        {
            mCollectors[mClosestCollector].activate();
            mActiveCollectors.push_back(&mCollectors[mClosestCollector]);
            playSoundEffect(SOUND_ACTIVATE);
        }
    }
}


void GameState::Ship::setTarget(FLOAT x, FLOAT y)
{
    targetX = x;
    targetY = y;
    if (goingToTarget)
        braking = TARGET_BRAKE_FRAMES;
    goingToTarget = TRUE;
}


void GameState::Collector::update()
{
    // no need to do this kind of sim here
    //FLOAT ox = x, oy = y;
    //x += DAMPING * (x - px);
    //y += DAMPING * (y - py);
    //px = ox; py = oy;
    if (active && runTimeLeft > 0)
    {
        --runTimeLeft;
        if (runTimeLeft < COLLECTOR_WARNING_AGE)
            angle += 0.1f / 60;
        else
            angle -= 0.5f / 60;
    }
}


void GameState::Ship::reset(FLOAT x, FLOAT y)
{
    this->x = x;
    this->y = y;
    angle = 0;
    vx = vy = 0;
    goingToTarget = FALSE;
    braking = 0;
    attachedCollector = -1;
    mLastBrakeSoundTick = 0;
}


void GameState::Ship::update(UINT32 time)
{
    x += vx;
    y += vy;
    vx *= DAMPING;
    vy *= DAMPING;

    FLOAT dx = targetX - x;
    FLOAT dy = targetY - y;
    FLOAT dsq = dx * dx + dy * dy;
    if (goingToTarget && dsq < TARGET_REACHED_DIST * TARGET_REACHED_DIST)
    {
        goingToTarget = FALSE;
        braking = TARGET_BRAKE_FRAMES;
        if (mLastBrakeSoundTick + 1000 < time)
        {
            playSoundEffect(SOUND_BRAKE);
            mLastBrakeSoundTick = time;
        }
    }

    if (braking > 0)
    {
        vx *= SHIP_BRAKE_MULT1;
        vy *= SHIP_BRAKE_MULT1;
        --braking;
        if (braking < TARGET_BRAKE_ENDING)
        {
            vx *= SHIP_BRAKE_MULT2;
            vy *= SHIP_BRAKE_MULT2;
        }
    }

    if (goingToTarget)
    {
        FLOAT bestDistSq = 1e10f;
        INT a, bestAngle = 1;
        for (a = 0; a < 3; ++a)
        {
            FLOAT testAngle = angle - SHIP_ROTATION_RATE + a * SHIP_ROTATION_RATE;
            FLOAT tx = (FLOAT)cos(testAngle);
            FLOAT ty = (FLOAT)sin(testAngle);
            FLOAT xd = dx * 1000 - tx, yd = dy * 1000 - ty;
            FLOAT dsq = xd * xd + yd * yd;
            if (dsq < bestDistSq)
            {
                bestDistSq = dsq;
                bestAngle = a;
            }
        }
        angle += (bestAngle - 1) * SHIP_ROTATION_RATE;

        FLOAT thrust = SHIP_THRUST;
        if (bestAngle != 1)
        {
            thrust *= 0.5f;
        }
        vx += cos(angle) * thrust;
        vy += sin(angle) * thrust;
    }
}


void GameState::Base::update()
{
    angle -= 0.2f / 60;
}
