/* Copyright 2006 Jetro Lauha - http://iki.fi/jetro/
 *
 * $Id: GameState.h,v 1.2 2006-04-30 22:11:53 tonic Exp $
 * $Revision: 1.2 $
 */

#ifndef GAMESTATE_H_INCL
#define GAMESTATE_H_INCL


#include <vector>


enum GAMESTATE_ENUM
{
    LEVEL_COUNT             = 3,
    PARTICLE_COUNT          = 5000,
    VISUAL_SCALE_BASE       = HEIGHT,
    SHIP_IMAGE_COUNT        = 64,
    COLLECTOR_NORMAL_AGE    = 4 * 60 * 60,
    COLLECTOR_WARNING_AGE   = (INT)(1.5 * 60 * 60)
};

#define DAMPING             0.9999f
//#define DAMPING_ATTACHED    0.999f
#define COLLECTOR_DAMPING   0.99f
#define COLLECTOR_RADIUS    0.2f
#define COLLECTOR_TOW_RADIUS        0.1f
#define COLLECTOR_GREATER_RADIUS    0.6f
#define PARTICLE_TOW_DISPLACE_RATIO 0.85f
#define SHIP_ROTATION_RATE  (2 * PI / 256)
#define SHIP_THRUST         0.00001f
#define SHIP_BRAKE_MULT1    0.975f
#define SHIP_BRAKE_MULT2    0.9f
#define SHIP_RADIUS         0.025f
#define TARGET_REACHED_DIST 0.1f
#define TARGET_BRAKE_FRAMES 60
#define TARGET_BRAKE_ENDING 10
#define BASE_RADIUS         0.2f
#define BASE_EAT_RADIUS     0.04f
#define BASE_ACTIVE_RADIUS  0.3f


class GameState : public State, public UIEventListener
{
public:

    GameState(UIGraphics *graphics);
    virtual ~GameState();

    // From State:

    virtual BOOL init();
    virtual void deinit();

    virtual BOOL update(UINT32 time);
    virtual void render(SDL_Surface *screen);

    virtual void onActivate(UINT32 time);

    virtual void onKeyDown(const SDLKey &key);
    virtual void onMouseDown(const SDL_MouseButtonEvent &event);
    virtual void onMouseUp(const SDL_MouseButtonEvent &event);
    virtual void onMouseMotion(const SDL_MouseMotionEvent &event);

    // From UIEventListener:

    virtual void onAction(UIComponent *component);

    // Local:

    void resetGame();
    void startLevel(INT level);
    void returnToTitleMenu();

protected:

    void updateView();
    void updateCollectors();
    void simulateParticles();
    void updateShip();

    void updateContextButtons();

    void drawBackground();
    void drawCollectors();
    void drawCollectorLights();
    void drawParticles();
    void drawShip();
    void drawBase();
    void drawBaseLights();
    void drawTexts();

    void centerMouse();

    enum STATE
    {
        STATE_BRIEFING = 0,
        STATE_RUNNING,
        STATE_FINISHED
    };

    INT32 mState;
    UINT32 mTime, mStartTime;
    UI mUI;
    UIGraphics *mGraphics;

    UILabel *mLevelLabel;
    UIButton *mContinueButton, *mGrabButton, *mDetachButton, *mActivateButton;

    UIImage *mButtonUpImage, *mButtonDownImage;
    UIImage *mSmallButtonUpImage, *mSmallButtonDownImage;
    UIImage *mBackgroundImage, *mColorsImage[LEVEL_COUNT];
    UIImage *mCollectorImage, *mFlareImage, *mBaseImage, *mFlameImage;
    UIImage *mShipImages[SHIP_IMAGE_COUNT];

    INT32 mMouseX, mMouseY;
    BOOL mMouseDown;
    INT32 mMouseDownX, mMouseDownY;

    FLOAT mWorldRadius;
    FLOAT mViewCenterX, mViewCenterY;
    FLOAT mViewCenterTargetX, mViewCenterTargetY;

    MersenneTwister mRandom;
    FLOAT mVisualScale;

    struct Particle
    {
        Particle() { x = y = vx = vy = 0; active = TRUE; }
        Particle(FLOAT x, FLOAT y) { this->x = x; this->y = y; vx = vy = 0; active = TRUE; }
        Particle(FLOAT x, FLOAT y, FLOAT vx, FLOAT vy) { this->x = x; this->y = y; this->vx = vx; this->vy = vy; active = TRUE; }
        FLOAT x, y;
        FLOAT vx, vy;
        BOOL active;
    };

    struct Collector
    {
        Collector(FLOAT x, FLOAT y, FLOAT angle, INT age, INT index)
        {
            this->x = /*px =*/ x;
            this->y = /*py =*/ y;
            this->index = index;
            active = FALSE;
            runTimeLeft = age;
            this->angle = angle;
            dx = dy = 0;
        }
        void activate() { if (runTimeLeft > 0) active = TRUE; }
        void update();
        FLOAT x, y;
        //FLOAT px, py;
        FLOAT dx, dy;
        FLOAT angle;
        BOOL active;
        INT32 runTimeLeft;
        INT index;
    };

    struct Ship : public Particle
    {
        Ship() { angle = 0; }
        void reset(FLOAT x, FLOAT y);
        void update(UINT32 time);
        void setTarget(FLOAT x, FLOAT y);
        void attachCollector(INT index) { attachedCollector = index; }
        void detachCollector() { attachedCollector = -1; }
        FLOAT angle;
        FLOAT targetX, targetY;
        BOOL goingToTarget;
        INT braking;
        INT attachedCollector;  // -1 if not attached, or index when attached
        UINT32 mLastBrakeSoundTick;
    };

    struct Base
    {
        Base() { x = y = angle = 0; active = FALSE; }
        void reset(FLOAT x, FLOAT y) { this->x = x; this->y = y; angle = 0; active = FALSE; }
        void update();
        FLOAT x, y;
        FLOAT angle;
        BOOL active;
    };

    typedef std::vector<Particle> ParticleArray;
    typedef std::vector<Collector> CollectorArray;
    typedef std::vector<Collector *> CollectorPointerArray;

    ParticleArray mParticles;
    CollectorArray mCollectors;
    CollectorPointerArray mActiveCollectors;
    Ship mShip;
    Base mBase;
    INT mClosestCollector;
    INT mNextCollectorIndex;
    INT mNextViewCenterCollectorIndex;
    UINT32 mHelpTextStartTick;

    BOOL mCollectorEnergyLeft;
    INT32 mLevel;
    INT32 mScore;
    INT32 mParticlesLeft;
    FLOAT mBackgroundAngle;
    BOOL mUsedGraduationExtraPoints;
    FLOAT mBaseSoundFrequency;
    FLOAT mBaseSoundVolume;
    UINT32 mLastTargetSoundTick;
};


#endif // !GAMESTATE_H_INCL
