/* Copyright 2006 Jetro Lauha - http://iki.fi/jetro/
 *
 * $Id: sound.cpp,v 1.2 2006-04-30 22:11:53 tonic Exp $
 * $Revision: 1.2 $
 */

#include "fmod.h"
#include "main.h"
#include "sound.h"


//#define DISABLE_SOUND


#define MUSICFILE   DATA_PREFIX "LD48-8.oxm"
#define MUSICVOLMUL 0.5f
#define SFXVOLMUL   1.0f


typedef struct
{
    INT type;
    UINT8 volume;
    char *filename;
    BOOL loop;
} SOUNDEFF;


static SOUNDEFF sampleData[] =
{
    { SOUND_ACTIVATE, 0xff, DATA_PREFIX "activate.ogg", FALSE },
    { SOUND_BASE, 0xff, DATA_PREFIX "base.ogg", TRUE },
    { SOUND_BRAKE, 0xff, DATA_PREFIX "brake.ogg", FALSE },
    { SOUND_BUTTON, 0xff, DATA_PREFIX "button.ogg", FALSE },
    { SOUND_DETACH, 0xff, DATA_PREFIX "detach.ogg", FALSE },
    { SOUND_GRAB, 0xff, DATA_PREFIX "grab.ogg", FALSE },
    { SOUND_LEVELDONE, 0xff, DATA_PREFIX "leveldone.ogg", FALSE },
    { SOUND_STARTLEVEL, 0xff, DATA_PREFIX "startlevel.ogg", FALSE },
    { SOUND_TARGET, 0xff, DATA_PREFIX "target.ogg", FALSE }
};
#define SAMPLES (sizeof(sampleData) / sizeof(sampleData[0]))


static BOOL soundInitialized = FALSE, soundEnabled = TRUE, musicEnabled = FALSE;
static FSOUND_SAMPLE *samples[SAMPLES];
static FMUSIC_MODULE *music;
static int musicChn = -1;
static float musicVolume = 1.0f;


static FSOUND_SAMPLE * loadSample(char *fileName, bool loop)
{
#ifndef DISABLE_SOUND
    return FSOUND_Sample_Load(FSOUND_FREE, fileName,
        loop ? FSOUND_LOOP_NORMAL : FSOUND_LOOP_OFF, 0, 0);
#else
    return NULL;
#endif // DISABLE_SOUND
}


static void loadSamples()
{
#ifndef DISABLE_SOUND
    int a;

    for (a = 0; a < SAMPLES; ++a)
    {
        samples[a] = loadSample(sampleData[a].filename, sampleData[a].loop);
        if (samples[a] == NULL)
        {
            errorMessage("Sample Load Error", "Error loading sample %s\nError: %s",
                         sampleData[a].filename, SDL_GetError());
            exit(1);
        }
    }
#endif // !DISABLE_SOUND
}


static void loadMusic()
{
#ifndef DISABLE_SOUND
    music = FMUSIC_LoadSong(MUSICFILE);
    if (music == NULL)
    {
        errorMessage("Music Load Error", "Error loading file %s\nError code: %d",
                     MUSICFILE, FSOUND_GetError());
        musicEnabled = 0;
    }
#endif // !DISABLE_SOUND
}


void initSound()
{
#ifndef DISABLE_SOUND
    soundInitialized = 1;
    if (!FSOUND_Init(44100, CHANNELS, 0))
    {
        errorMessage("Sound Disabled", "Unable to initialize sound! (%d)",
                      FSOUND_GetError());
        soundInitialized = soundEnabled = 0;
        return;
    }

    loadSamples();
    loadMusic();
#endif // !DISABLE_SOUND
}


void deinitSound()
{
#ifndef DISABLE_SOUND
    if (!soundInitialized)
        return;

    setMusic(0, 0);

    UINT a;
    for (a = 0; a < SAMPLES; ++a)
    {
        FSOUND_Sample_Free(samples[a]);
        samples[a] = NULL;
    }
    FMUSIC_FreeSong(music);

    FSOUND_Close();
#endif // !DISABLE_SOUND
}


int playSoundEffect(int type, int forceChannel)
{
#ifndef DISABLE_SOUND
    int chn, snd, cnt = 0;

    if (!soundInitialized || !soundEnabled)
        return -1;

    do {
        snd = rand() % SAMPLES;
        ++cnt;
    } while (sampleData[snd].type != type && cnt < 10000);
    if (cnt >= 10000)
        return -1; // don't play anything if we can't find a sample
        //errorMessage("Sound Error", "Unable to find sample");

    if (forceChannel >= 0)
        chn = FSOUND_PlaySound(forceChannel, samples[snd]);
    else
        chn = FSOUND_PlaySound(FSOUND_FREE, samples[snd]);

    if (chn != -1)
        FSOUND_SetVolume(chn, sampleData[snd].volume);

    return chn;
#else
    return 0;
#endif // DISABLE_SOUND
}


void stopSoundEffect(int channel)
{
#ifndef DISABLE_SOUND
    if (!soundInitialized || channel < 0)
        return;
    FSOUND_StopSound(channel);
#endif // DISABLE_SOUND
}


void stopAllSoundEffects()
{
#ifndef DISABLE_SOUND
    if (!soundInitialized)
        return;

    int a;
    for (a = 0; a < CHANNELS; ++a)
    {
        if (musicEnabled && musicChn == a)
            continue;   // don't stop music if music is enabled
        stopSoundEffect(a);
    }
#endif // DISABLE_SOUND
}


void setSoundChannelVolume(int channel, int volume)
{
#ifndef DISABLE_SOUND
    if (volume < 0)
        volume = 0;
    if (volume > 0xff)
        volume = 0xff;
    if (channel != -1)
        FSOUND_SetVolume(channel, volume);
#endif // DISABLE_SOUND
}


void setSoundChannelFrequency(int channel, int freq)
{
#ifndef DISABLE_SOUND
    if (freq < 1000)
        freq = 1000;
    if (freq > 500000)
        freq = 500000;
    if (channel != -1)
        FSOUND_SetFrequency(channel, freq);
#endif // DISABLE_SOUND
}


void setSound(BOOL enabled)
{
#ifndef DISABLE_SOUND
    if (!soundInitialized)
        return;

    soundEnabled = enabled ? 1 : 0;
#endif // !DISABLE_SOUND
}


void setMusic(BOOL enabled, INT order)
{
#ifndef DISABLE_SOUND
    if (!soundInitialized)
        return;

    if (enabled && !musicEnabled)
    {
        // turn music on
        musicEnabled = FMUSIC_PlaySong(music) ? 1 : 0;
        setMusicVolume(musicVolume);
    }
    else if (musicEnabled && !enabled)
    {
        // turn music off
        FMUSIC_StopAllSongs();
        musicEnabled = 0;
    }
    if (musicEnabled)
        FMUSIC_SetOrder(music, order);

#endif // !DISABLE_SOUND
}


void setMusicVolume(float volume)
{
#ifndef DISABLE_SOUND
    if (!soundInitialized)
        return;

    musicVolume = volume;

    if (musicEnabled)
        FMUSIC_SetMasterVolume(music, (int)(volume * MUSICVOLMUL * 256));
#endif // !DISABLE_SOUND
}


BOOL getSoundEnabled()
{
#ifndef DISABLE_SOUND
    return soundEnabled;
#else
    return FALSE;
#endif
}


BOOL getMusicEnabled()
{
#ifndef DISABLE_SOUND
    return musicEnabled;
#else
    return FALSE;
#endif
}
