/* Turska UI library
 * Copyright 2003-2004 Jetro Lauha
 * All rights reserved.
 * Web: http://iki.fi/jetro/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of EITHER:
 *   (1) The GNU Lesser General Public License as published by the Free
 *       Software Foundation; either version 2.1 of the License, or (at
 *       your option) any later version. The text of the GNU Lesser
 *       General Public License is included with this library in the
 *       file LICENSE-LGPL.txt.
 *   (2) The BSD-style license that is included with this library in
 *       the file LICENSE-BSD.txt.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the files
 * LICENSE-LGPL.txt and LICENSE-BSD.txt for more details.
 *
 * $Id: UIFont.h,v 1.2 2006-04-30 14:06:38 tonic Exp $
 * $Revision: 1.2 $
 */

#ifndef UIFONT_H_INCLUDED
#define UIFONT_H_INCLUDED


class UIFont
{
public:

    //! Font alignment flags.
    /*! Default alignment value of 0 implies top vertical alignment and left horizontal alignment.
     */
    enum ALIGN_FLAG
    {
        ALIGN_HORIZ_CENTER = (1 << 0),
        ALIGN_RIGHT        = (1 << 1),
        ALIGN_VERT_CENTER  = (1 << 2),
        ALIGN_BOTTOM       = (1 << 3)
    };

    //! Defines UI font related constants.
    enum CONSTANTS
    {
        //! Maximum file name length.
        MAX_FILENAME_LENGTH = 256,
        //! Maximum amount of glyphs - current implementation assumes 8 bits per character.
        MAX_GLYPHS          = 256   
    };

    //! Constructor caches the file name and builds glyph map using font string.
    /*! \note Default font string is
     *        " !\"#$%&'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]^_`abcdefghijklmnopqrstuvwxyz{|}~".
     * \param fileName image file to load the font from.
     * \param fontString ordered string of characters appearing in the font image.
     */
    UIFont(const char *fileName, const char *fontString = NULL);
    virtual ~UIFont();

    //! Does any initialization the object needs before it is ready for use.
    /*!
     * \return TRUE if successful or FALSE otherwise.
     */
    virtual BOOL init() = 0;

    //! Returns TRUE if the font is loaded or FALSE otherwise.
    BOOL isLoaded() { return mLoaded; }

    void setCharacterSpacing(INT spacing) { mCharacterSpacing = spacing; }
    INT getCharacterSpacing() { return mCharacterSpacing; }
    UINT getFontHeight() { return mFontHeight; }

    //! Reloads the font.
    /*! Sets the internal loaded state to TRUE if loading was successful and FALSE otherwise.
     * \return TRUE if successful or FALSE otherwise.
     */
    virtual BOOL reload() = 0;

    virtual INT32 getStringWidth(const char *string) = 0;

    //! Sets draw mode for glyph images. \sa UIImage::DRAW_MODE
    void setDrawMode(UINT drawMode) { mDrawMode = drawMode; }

    //! Draws a string.
    /*!
     * \param x X-coordinate of the origin for the string.
     * \param y Y-coordinate of the origin for the string.
     * \param flags flags to control how the string is positioned and drawn. \sa ALIGN_FLAG
     * \param string string to draw.
     * \param red red color component.
     * \param green green color component.
     * \param blue blue color component.
     * \param alpha alpha component.
     */
    virtual void drawString(FLOAT x, FLOAT y, UINT32 flags, const char *string,
                            FLOAT red = 1, FLOAT green = 1, FLOAT blue = 1, FLOAT alpha = 1) = 0;

    //! Draws a string.
    /*! The color parameter array should have 4 floats with values for red, green, blue and alpha respectively.
     */
    virtual void drawString(FLOAT x, FLOAT y, UINT32 flags, const char *string, const FLOAT *color) = 0;

protected:

    BOOL mLoaded;
    char mFileName[MAX_FILENAME_LENGTH];
    UINT8 mGlyphMap[MAX_GLYPHS];
    UIImage **mGlyphs;
    UINT mGlyphCount;
    INT mCharacterSpacing;
    UINT mFontHeight;
    UINT mDrawMode;

    void buildCharMap(const char *fontString);
};


#endif // !UIFONT_H_INCLUDED
