/* Turska UI library
 * Copyright 2003-2004 Jetro Lauha
 * All rights reserved.
 * Web: http://iki.fi/jetro/ or http://jet.ro
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of EITHER:
 *   (1) The GNU Lesser General Public License as published by the Free
 *       Software Foundation; either version 2.1 of the License, or (at
 *       your option) any later version. The text of the GNU Lesser
 *       General Public License is included with this library in the
 *       file LICENSE-LGPL.txt.
 *   (2) The BSD-style license that is included with this library in
 *       the file LICENSE-BSD.txt.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the files
 * LICENSE-LGPL.txt and LICENSE-BSD.txt for more details.
 *
 * $Id: UIGraphicsOpenGLSDL.cpp,v 1.2 2006-04-30 14:06:34 tonic Exp $
 * $Revision: 1.2 $
 */

#include "SDL.h"
#include "SDL_opengl.h"
#include "SDL_image.h"
#include "UI.h"


static inline UINT32 nextPowerOfTwo(UINT32 input)
{
     --input;
     input |= input >> 16;
     input |= input >> 8;
     input |= input >> 4;
     input |= input >> 2;
     input |= input >> 1;
     return input + 1;
}


// Based on SDL examples.
static void SDL_GL_Enter2DMode()
{
    SDL_Surface *screen = SDL_GetVideoSurface();

    glPushAttrib(GL_COLOR_BUFFER_BIT);
    glPushAttrib(GL_ENABLE_BIT);
    glDisable(GL_DEPTH_TEST);
    glDisable(GL_CULL_FACE);
    glEnable(GL_TEXTURE_2D);

    // This allows alpha blending of 2D textures with the scene.
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

    glViewport(0, 0, screen->w, screen->h);

    glMatrixMode(GL_PROJECTION);
    glPushMatrix();
    glLoadIdentity();

    glOrtho(0, (GLdouble)screen->w, (GLdouble)screen->h, 0, 0, 1);

    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
    glLoadIdentity();

    glTranslatef(0.375f, 0.375f, 0);
 
    //glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_DECAL);
    glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
    glColor4f(1, 1, 1, 1);
}


// Based on SDL examples.
static void SDL_GL_Leave2DMode()
{
    glMatrixMode(GL_MODELVIEW);
    glPopMatrix();

    glMatrixMode(GL_PROJECTION);
    glPopMatrix();

    glPopAttrib(); // GL_ENABLE_BIT
    glPopAttrib(); // GL_COLOR_BUFFER_BIT
}


void UIGraphicsOpenGLSDL::beginRender()
{
    SDL_GL_Enter2DMode();
}


void UIGraphicsOpenGLSDL::endRender()
{
    SDL_GL_Leave2DMode();
}



UIImage * UIGraphicsOpenGLSDL::loadImage(const char *fileName, UINT32 flags)
{
    SDL_RWops *rw = SDL_RWFromFile(fileName, "rb");
    if (rw == NULL)
        return NULL;
    return loadImage(rw, TRUE, flags);
}


UIImage * UIGraphicsOpenGLSDL::loadImage(SDL_RWops *source, BOOL freeSource,
                                         UINT32 flags)
{
    SDL_Surface *surface;

    surface = loadSurface(source, freeSource);
    if (surface == NULL)
        return NULL;

    UIImage *res = createImage(surface, flags);
    SDL_FreeSurface(surface);

    return res;
}


SDL_Surface * UIGraphicsOpenGLSDL::loadSurface(const char *fileName)
{
    SDL_RWops *rw = SDL_RWFromFile(fileName, "rb");
    if (rw == NULL)
        return NULL;
    return loadSurface(rw, TRUE);
}


SDL_Surface * UIGraphicsOpenGLSDL::loadSurface(SDL_RWops *source, BOOL deleteSource)
{
    return IMG_Load_RW(source, deleteSource ? 1 : 0);
}


UIImage * UIGraphicsOpenGLSDL::createImage(SDL_Surface *surface, UINT32 flags,
                                           UIRectangle *sourceRect)
{
    BOOL success;
    GLuint texture;
    GLfloat uv[4];

    if (surface == NULL)
        return NULL;

    success = createTexture(texture, surface, flags, uv, sourceRect);
    if (!success)
        return NULL;

    if (sourceRect == NULL)
        return new UIImageOpenGL(surface->w, surface->h, flags, texture, uv);
    else
        return new UIImageOpenGL(sourceRect->w, sourceRect->h, flags, texture, uv);
}


static void sSimpleFixAlphaChannel(SDL_Surface *image)
{
    if (image->w < 4 || image->h < 4)
        return;

    if (SDL_LockSurface(image) != 0)
        return;

    Uint32 *pixels = (Uint32 *)image->pixels;
    INT32 x, y;
    INT32 w = image->w, h = image->h;
    const UINT32 alphaMask = image->format->Amask;
    const UINT32 colorMask =
        image->format->Rmask | image->format->Gmask | image->format->Bmask;

    for (y = h - 2; y >= 0; --y)
    {
        for (x = w - 2; x >= 0; --x)
        {
            if ((pixels[y * w + x] & alphaMask) == 0)
                pixels[y * w + x] = pixels[(y + 1) * w + x + 1] & colorMask;
        }
    }

    for (y = h - 2; y >= 0; --y)
    {
        for (x = 1; x < w; ++x)
        {
            if ((pixels[y * w + x] & alphaMask) == 0)
                pixels[y * w + x] = pixels[(y + 1) * w + x - 1] & colorMask;
        }
    }

    for (y = 1; y < h; ++y)
    {
        for (x = w - 2; x >= 0; --x)
        {
            if ((pixels[y * w + x] & alphaMask) == 0)
                pixels[y * w + x] = pixels[(y - 1) * w + x + 1] & colorMask;
        }
    }

    for (y = 1; y < h; ++y)
    {
        for (x = 1; x < w; ++x)
        {
            if ((pixels[y * w + x] & alphaMask) == 0)
                pixels[y * w + x] = pixels[(y - 1) * w + x - 1] & colorMask;
        }
    }

    SDL_UnlockSurface(image);
}


// Based on SDL examples.
BOOL UIGraphicsOpenGLSDL::createTexture(UINT &texture, SDL_Surface *surface, UINT32 flags, FLOAT *uv, UIRectangle *sourceRect)
{
    UINT32 w, h;
    SDL_Surface *image;
    SDL_Rect area;
    SDL_Rect srcr;
    UINT32 savedFlags;
    UINT8 savedAlpha;
    UINT srcw, srch;

    if (sourceRect != NULL)
    {
        srcw = sourceRect->w;
        srch = sourceRect->h;
    }
    else
    {
        srcw = surface->w;
        srch = surface->h;
    }

    // Use the surface width and height expanded to powers of 2.
    w = nextPowerOfTwo(srcw);
    h = nextPowerOfTwo(srch);
    /*
    uv[0] = 0.5f / w;                  // Min X
    uv[1] = 0.5f / h;                  // Min Y
    uv[2] = (GLfloat)srcw / w;  // Max X
    uv[3] = (GLfloat)srch / h;  // Max Y
    */
    uv[0] = 0;                  // Min X
    uv[1] = 0;                  // Min Y
    uv[2] = ((GLfloat)srcw - 0.0001f) / w;  // Max X
    uv[3] = ((GLfloat)srch - 0.0001f) / h;  // Max Y

    image = SDL_CreateRGBSurface(
        SDL_HWSURFACE | SDL_SRCALPHA,
        w, h, 32,
#if SDL_BYTEORDER == SDL_LIL_ENDIAN /* OpenGL RGBA masks */
        0x000000FF, 
        0x0000FF00, 
        0x00FF0000, 
        0xFF000000
#else
        0xFF000000,
        0x00FF0000, 
        0x0000FF00, 
        0x000000FF
#endif
        );
    if (image == NULL)
        return FALSE;

    // Save the alpha blending attributes.
    savedFlags = surface->flags & (SDL_SRCALPHA | SDL_RLEACCELOK);
    savedAlpha = surface->format->alpha;
    if ((savedFlags & SDL_SRCALPHA) == SDL_SRCALPHA)
        SDL_SetAlpha(surface, 0, 0);

    // Copy the surface into the GL texture image.
    area.x = 0;
    area.y = 0;
    area.w = srcw;
    area.h = srch;
    if (sourceRect != NULL)
    {
        srcr.x = sourceRect->x;
        srcr.y = sourceRect->y;
        srcr.w = sourceRect->w;
        srcr.h = sourceRect->h;
    }
    else
        srcr = area;
    SDL_BlitSurface(surface, &srcr, image, &area);

    // TODO: Should this "fix" be optional?
    sSimpleFixAlphaChannel(image);

    // Restore the alpha blending attributes.
    if ((savedFlags & SDL_SRCALPHA) == SDL_SRCALPHA)
        SDL_SetAlpha(surface, savedFlags, savedAlpha);

    // Create an OpenGL texture from the image.
    glGenTextures(1, &texture);
    glBindTexture(GL_TEXTURE_2D, texture);

    if (flags & UIImage::FILTER_LINEAR)
    {
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    }
    else
    {
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
    }

    //glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
    //glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);

    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, w, h, 0, GL_RGBA, GL_UNSIGNED_BYTE, image->pixels);
    SDL_FreeSurface(image); // No longer needed.

    return TRUE;
}
